@@ -26,6 +26,7 @@ License along with this library. If not, see
 #endif
 
 #include <glib/gi18n-lib.h>
+#include <gio/gio.h>
 
 #include "device.h"
 
@@ -301,16 +302,33 @@ get_device_icon_index (gdouble percentage)
   return "000";
 }
 
+static const char *
+device_kind_to_string (UpDeviceKind kind)
+{
+  switch (kind)
+    {
+      case UP_DEVICE_KIND_LINE_POWER: return "line-power";
+      case UP_DEVICE_KIND_BATTERY: return "battery";
+      case UP_DEVICE_KIND_UPS: return "ups";
+      case UP_DEVICE_KIND_MONITOR: return "monitor";
+      case UP_DEVICE_KIND_MOUSE: return "mouse";
+      case UP_DEVICE_KIND_KEYBOARD: return "keyboard";
+      case UP_DEVICE_KIND_PDA: return "pda";
+      case UP_DEVICE_KIND_PHONE: return "phone";
+      case UP_DEVICE_KIND_MEDIA_PLAYER: return "media-player";
+      case UP_DEVICE_KIND_TABLET: return "tablet";
+      case UP_DEVICE_KIND_COMPUTER: return "computer";
+      default: return "unknown";
+    }
+}
+
 /**
   indicator_power_device_get_icon_names:
   @device: #IndicatorPowerDevice from which to generate the icon names
 
   This function's logic differs from GSD's power plugin in some ways:
 
-  1. All charging batteries use the same icon regardless of progress.
-  <https://bugs.launchpad.net/indicator-power/+bug/824629/comments/7>
-
-  2. For discharging batteries, we decide whether or not to use the 'caution'
+  1. For discharging batteries, we decide whether or not to use the 'caution'
   icon based on whether or not we have <= 30 minutes remaining, rather than
   looking at the battery's percentage left.
   <https://bugs.launchpad.net/indicator-power/+bug/743823>
@@ -334,7 +352,7 @@ indicator_power_device_get_icon_names (const IndicatorPowerDevice * device)
   gdouble percentage = indicator_power_device_get_percentage (device);
   const UpDeviceKind kind = indicator_power_device_get_kind (device);
   const UpDeviceState state = indicator_power_device_get_state (device);
-  const gchar * kind_str = kind_str = up_device_kind_to_string (kind);
+  const gchar * kind_str = device_kind_to_string (kind);
 
   GPtrArray * names = g_ptr_array_new ();
 
@@ -351,32 +369,28 @@ indicator_power_device_get_icon_names (const IndicatorPowerDevice * device)
   else switch (state)
     {
       case UP_DEVICE_STATE_EMPTY:
-        g_ptr_array_add (names, g_strdup("battery-empty-symbolic"));
+        g_ptr_array_add (names, g_strdup_printf("%s-empty-symbolic", kind_str));
         g_ptr_array_add (names, g_strdup_printf("gpm-%s-empty", kind_str));
         g_ptr_array_add (names, g_strdup_printf("gpm-%s-000", kind_str));
-        g_ptr_array_add (names, g_strdup("battery-empty"));
+        g_ptr_array_add (names, g_strdup_printf("%s-empty", kind_str));
         break;
 
       case UP_DEVICE_STATE_FULLY_CHARGED:
-        g_ptr_array_add (names, g_strdup("battery-full-charged-symbolic"));
-        g_ptr_array_add (names, g_strdup("battery-full-charging-symbolic"));
+        g_ptr_array_add (names, g_strdup_printf("%s-full-charged-symbolic", kind_str));
+        g_ptr_array_add (names, g_strdup_printf("%s-full-charging-symbolic", kind_str));
         g_ptr_array_add (names, g_strdup_printf("gpm-%s-full", kind_str));
         g_ptr_array_add (names, g_strdup_printf("gpm-%s-100", kind_str));
-        g_ptr_array_add (names, g_strdup("battery-full-charged"));
-        g_ptr_array_add (names, g_strdup("battery-full-charging"));
+        g_ptr_array_add (names, g_strdup_printf("%s-full-charged", kind_str));
+        g_ptr_array_add (names, g_strdup_printf("%s-full-charging", kind_str));
         break;
 
       case UP_DEVICE_STATE_CHARGING:
       case UP_DEVICE_STATE_PENDING_CHARGE:
-        /* When charging, always use the same icon regardless of percentage.
-           <http://bugs.launchpad.net/indicator-power/+bug/824629> */
-        percentage = 0;
-
         suffix_str = get_device_icon_suffix (percentage);
         index_str = get_device_icon_index (percentage);
-        g_ptr_array_add (names, g_strdup_printf ("battery-%s-charging-symbolic", suffix_str));
+        g_ptr_array_add (names, g_strdup_printf ("%s-%s-charging-symbolic", kind_str, suffix_str));
         g_ptr_array_add (names, g_strdup_printf ("gpm-%s-%s-charging", kind_str, index_str));
-        g_ptr_array_add (names, g_strdup_printf ("battery-%s-charging", suffix_str));
+        g_ptr_array_add (names, g_strdup_printf ("%s-%s-charging", kind_str, suffix_str));
         break;
 
       case UP_DEVICE_STATE_DISCHARGING:
@@ -392,14 +406,14 @@ indicator_power_device_get_icon_names (const IndicatorPowerDevice * device)
         index_str = get_device_icon_index (percentage);
         g_ptr_array_add (names, g_strdup_printf ("%s-%s", kind_str, index_str));
         g_ptr_array_add (names, g_strdup_printf ("gpm-%s-%s", kind_str, index_str));
-        g_ptr_array_add (names, g_strdup_printf ("battery-%s-symbolic", suffix_str));
-        g_ptr_array_add (names, g_strdup_printf ("battery-%s", suffix_str));
+        g_ptr_array_add (names, g_strdup_printf ("%s-%s-symbolic", kind_str, suffix_str));
+        g_ptr_array_add (names, g_strdup_printf ("%s-%s", kind_str, suffix_str));
         break;
 
       default:
-        g_ptr_array_add (names, g_strdup("battery-missing-symbolic"));
-        g_ptr_array_add (names, g_strdup("gpm-battery-missing"));
-        g_ptr_array_add (names, g_strdup("battery-missing"));
+        g_ptr_array_add (names, g_strdup_printf("%s-missing-symbolic", kind_str));
+        g_ptr_array_add (names, g_strdup_printf("gpm-%s-missing", kind_str));
+        g_ptr_array_add (names, g_strdup_printf("%s-missing", kind_str));
     }
 
     g_ptr_array_add (names, NULL); /* terminates the strv */
@@ -428,10 +442,11 @@ indicator_power_device_get_gicon (const IndicatorPowerDevice * device)
 ****
 ***/
 
+/* Format time remaining for reading ("H:MM") and speech ("H hours, MM minutes") */
 static void
 get_timestring (guint64   time_secs,
-                gchar   **short_timestring,
-                gchar   **detailed_timestring)
+                gchar   **readable_timestring,
+                gchar   **accessible_timestring)
 {
   gint  hours;
   gint  minutes;
@@ -441,16 +456,16 @@ get_timestring (guint64   time_secs,
 
   if (minutes == 0)
     {
-      *short_timestring = g_strdup (_("Unknown time"));
-      *detailed_timestring = g_strdup (_("Unknown time"));
+      *readable_timestring = g_strdup (_("Unknown time"));
+      *accessible_timestring = g_strdup (_("Unknown time"));
 
       return;
     }
 
   if (minutes < 60)
     {
-      *short_timestring = g_strdup_printf ("0:%.2i", minutes);
-      *detailed_timestring = g_strdup_printf (g_dngettext (GETTEXT_PACKAGE, "%i minute",
+      *readable_timestring = g_strdup_printf ("0:%.2i", minutes);
+      *accessible_timestring = g_strdup_printf (g_dngettext (GETTEXT_PACKAGE, "%i minute",
                                               "%i minutes",
                                               minutes), minutes);
       return;
@@ -459,11 +474,11 @@ get_timestring (guint64   time_secs,
   hours = minutes / 60;
   minutes = minutes % 60;
 
-  *short_timestring = g_strdup_printf ("%i:%.2i", hours, minutes);
+  *readable_timestring = g_strdup_printf ("%i:%.2i", hours, minutes);
 
   if (minutes == 0)
     {
-      *detailed_timestring = g_strdup_printf (g_dngettext (GETTEXT_PACKAGE, 
+      *accessible_timestring = g_strdup_printf (g_dngettext (GETTEXT_PACKAGE,
                                               "%i hour",
                                               "%i hours",
                                               hours), hours);
@@ -472,7 +487,7 @@ get_timestring (guint64   time_secs,
     {
       /* TRANSLATOR: "%i %s %i %s" are "%i hours %i minutes"
        * Swap order with "%2$s %2$i %1$s %1$i if needed */
-      *detailed_timestring = g_strdup_printf (_("%i %s %i %s"),
+      *accessible_timestring = g_strdup_printf (_("%i %s %i %s"),
                                               hours, g_dngettext (GETTEXT_PACKAGE, "hour", "hours", hours),
                                               minutes, g_dngettext (GETTEXT_PACKAGE, "minute", "minutes", minutes));
     }
@@ -528,94 +543,161 @@ device_kind_to_localised_string (UpDeviceKind kind)
       /* TRANSLATORS: tablet device */
       text = _("Computer");
       break;
+    case UP_DEVICE_KIND_UNKNOWN:
+      /* TRANSLATORS: unknown device */
+      text = _("Unknown");
+      break;
     default:
       g_warning ("enum unrecognised: %i", kind);
-      text = up_device_kind_to_string (kind);
+      text = device_kind_to_string (kind);
     }
 
   return text;
 }
 
-void
-indicator_power_device_get_time_details (const IndicatorPowerDevice * device,
-                                         gchar ** short_details,
-                                         gchar ** details,
-                                         gchar ** accessible_name)
+static char *
+join_strings (const char * name, const char * time, const char * percent)
+{
+  char * str;
+  const gboolean have_name = name && *name;
+  const gboolean have_time = time && *time;
+  const gboolean have_percent = percent && *percent;
+
+  if (have_name && have_time && have_percent)
+    str = g_strdup_printf (_("%s (%s, %s)"), name, time, percent);
+  else if (have_name && have_time)
+    str = g_strdup_printf (_("%s (%s)"), name, time);
+  else if (have_name && have_percent)
+    str = g_strdup_printf (_("%s (%s)"), name, percent);
+  else if (have_name)
+    str = g_strdup (name);
+  else if (have_time && have_percent)
+    str = g_strdup_printf (_("(%s, %s)"), time, percent);
+  else if (have_time)
+    str = g_strdup_printf (_("(%s)"), time);
+  else if (have_percent)
+    str = g_strdup_printf (_("(%s)"), percent);
+  else
+    str = g_strdup ("");
+
+  return str;
+}
+
+static void
+indicator_power_device_get_text (const IndicatorPowerDevice * device,
+                                 gboolean show_time_in_header,
+                                 gboolean show_percentage_in_header,
+                                 gchar ** header,
+                                 gchar ** label,
+                                 gchar ** a11y)
 {
   if (!INDICATOR_IS_POWER_DEVICE(device))
     {
-      *short_details = NULL;
-      *details = NULL;
-      *accessible_name = NULL;
+      if (a11y != NULL) *a11y = NULL;
+      if (label != NULL) *label = NULL;
+      if (header != NULL) *header = NULL;
       g_warning ("%s: %p is not an IndicatorPowerDevice", G_STRFUNC, device);
       return;
     }
 
   const time_t time = indicator_power_device_get_time (device);
   const UpDeviceState state = indicator_power_device_get_state (device);
-  const gdouble percentage = indicator_power_device_get_percentage (device);
   const UpDeviceKind kind = indicator_power_device_get_kind (device);
   const gchar * device_name = device_kind_to_localised_string (kind);
+  const gdouble percentage = indicator_power_device_get_percentage (device);
+  char pctstr[32] = { '\0' };
+  g_snprintf (pctstr, sizeof(pctstr), "%.0lf%%", percentage);
+
+  GString * terse_time = g_string_new (NULL);
+  GString * verbose_time = g_string_new (NULL);
+  GString * accessible_time = g_string_new (NULL);
 
   if (time > 0)
     {
-      gchar *short_timestring = NULL;
-      gchar *detailed_timestring = NULL;
-
-      get_timestring (time,
-                      &short_timestring,
-                      &detailed_timestring);
+      char * readable_timestr = NULL;
+      char * accessible_timestr = NULL;
+      get_timestring (time, &readable_timestr, &accessible_timestr);
 
       if (state == UP_DEVICE_STATE_CHARGING)
         {
-          /* TRANSLATORS: %2 is a time string, e.g. "1 hour 5 minutes" */
-          *accessible_name = g_strdup_printf (_("%s (%s to charge (%.0lf%%))"), device_name, detailed_timestring, percentage);
-          *details = g_strdup_printf (_("%s (%s to charge)"), device_name, short_timestring);
-          *short_details = g_strdup_printf ("(%s)", short_timestring);
+          g_string_assign (terse_time, readable_timestr);
+          g_string_printf (verbose_time, _("%s to charge"), readable_timestr);
+          g_string_printf (accessible_time, _("%s to charge"), accessible_timestr);
         }
-      else if ((state == UP_DEVICE_STATE_DISCHARGING) && (time > (60*60*12)))
+      else if ((state == UP_DEVICE_STATE_DISCHARGING) && (time <= (60*60*24)))
         {
-          *accessible_name = g_strdup_printf (_("%s"), device_name);
-          *details = g_strdup_printf (_("%s"), device_name);
-          *short_details = g_strdup (short_timestring);
+          g_string_assign (terse_time, readable_timestr);
+          g_string_printf (verbose_time, _("%s left"), readable_timestr);
+          g_string_printf (accessible_time, _("%s left"), accessible_timestr);
         }
       else
         {
-          /* TRANSLATORS: %2 is a time string, e.g. "1 hour 5 minutes" */
-          *accessible_name = g_strdup_printf (_("%s (%s left (%.0lf%%))"), device_name, detailed_timestring, percentage);
-          *details = g_strdup_printf (_("%s (%s left)"), device_name, short_timestring);
-          *short_details = g_strdup (short_timestring);
+          /* if there's more than 24 hours remaining, we don't show it */
         }
 
-      g_free (short_timestring);
-      g_free (detailed_timestring);
+      g_free (readable_timestr);
+      g_free (accessible_timestr);
     }
   else if (state == UP_DEVICE_STATE_FULLY_CHARGED)
     {
-      *details = g_strdup_printf (_("%s (charged)"), device_name);
-      *accessible_name = g_strdup (*details);
-      *short_details = g_strdup ("");
+      g_string_assign (verbose_time,    _("charged"));
+      g_string_assign (accessible_time, _("charged"));
     }
   else if (percentage > 0)
     {
-      /* TRANSLATORS: %2 is a percentage value. Note: this string is only
-       * used when we don't have a time value */
-      *details = g_strdup_printf (_("%s (%.0lf%%)"), device_name, percentage);
-      *accessible_name = g_strdup (*details);
-      *short_details = g_strdup_printf (_("(%.0lf%%)"), percentage);
-    }
-  else if (kind == UP_DEVICE_KIND_LINE_POWER)
-    {
-      *details         = g_strdup (device_name);
-      *accessible_name = g_strdup (device_name);
-      *short_details   = g_strdup (device_name);
+      g_string_assign (terse_time,      _("estimating…"));
+      g_string_assign (verbose_time,    _("estimating…"));
+      g_string_assign (accessible_time, _("estimating…"));
     }
   else
     {
-      *details = g_strdup_printf (_("%s (not present)"), device_name);
-      *accessible_name = g_strdup (*details);
-      *short_details = g_strdup (_("(not present)"));
+      *pctstr = '\0';
+
+      if (kind != UP_DEVICE_KIND_LINE_POWER)
+        {
+          g_string_assign (verbose_time,    _("not present"));
+          g_string_assign (accessible_time, _("not present"));
+        }
     }
+
+  if (header != NULL)
+    *header = join_strings (NULL,
+                            show_time_in_header ? terse_time->str : "",
+                            show_percentage_in_header ? pctstr : "");
+
+  if (label != NULL)
+    *label  = join_strings (device_name,
+                            verbose_time->str,
+                            NULL);
+
+  if (a11y != NULL)
+    *a11y   = join_strings (device_name,
+                            accessible_time->str,
+                            pctstr);
+
+  g_string_free (terse_time, TRUE);
+  g_string_free (verbose_time, TRUE);
+  g_string_free (accessible_time, TRUE);
+}
+
+gchar *
+indicator_power_device_get_label (const IndicatorPowerDevice * device)
+{
+  gchar * label = NULL;
+  indicator_power_device_get_text (device, FALSE, FALSE,
+                                   NULL, &label, NULL);
+  return label;
+}
+
+void
+indicator_power_device_get_header (const IndicatorPowerDevice * device,
+                                   gboolean                     show_time,
+                                   gboolean                     show_percentage,
+                                   gchar                     ** header,
+                                   gchar                     ** a11y)
+{
+  indicator_power_device_get_text (device, show_time, show_percentage,
+                                   header, NULL, a11y);
 }
 
 /***
